# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import json
import os

import ecmwfapi

from climetlab.decorators import parameters

from .base import APIKeyPrompt, FileSource


class MARSAPI(APIKeyPrompt):

    text_message = """
An API key is needed to access this dataset. Please visit
https://apps.ecmwf.int/registration/ to register or sign-in
at https://www.ecmwf.int/user/login/sso
then visit https://api.ecmwf.int/v1/key/ to
retrieve you API key.

Once this is done, please copy the text that look like:

{
    "url"   : "https://api.ecmwf.int/v1",
    "key"   : "xxxxxxxxxxxxxxxxxxxxxxxxxxx",
    "email" : "john.doe@example.com"
}

paste it the input field below and press ENTER.
"""

    markdown_message = """
An API key is needed to access this dataset. Please visit
<https://apps.ecmwf.int/registration/> to register or sign-in
at <https://www.ecmwf.int/user/login/sso>
then visit <https://api.ecmwf.int/v1/key/> to
retrieve you API key.

Once this is done, please copy the text that look like:

<pre>
{
    "url"   : "https://api.ecmwf.int/v1",
    "key"   : "xxxxxxxxxxxxxxxxxxxxxxxxxxx",
    "email" : "john.doe@example.com"
}
</pre>

paste it the input field below and press *ENTER*.
"""

    rcfile = "~/.ecmwfapirc"
    prompt = "ECMWF api configuration"

    def validate(self, text):
        return json.dumps(json.loads(text), indent=4)


def service(name):
    try:
        return ecmwfapi.ECMWFService(name)
    except Exception as e:
        if ".ecmwfapirc" in str(e):
            MARSAPI().ask_user_and_save()
            return ecmwfapi.ECMWFService(name)

        raise


class MARSRetriever(FileSource):
    def __init__(self, **kwargs):
        request = self.request(**kwargs)
        self.path = self.cache_file(request)
        if not os.path.exists(self.path):
            service("mars").execute(request, self.path + ".tmp")
            os.rename(self.path + ".tmp", self.path)

    @parameters(
        param=("parameter-list", "mars"),
        date=("date-list", "%Y-%m-%d"),
        area=("bounding-box", list),
    )
    def request(self, **kwargs):
        return kwargs

    def read_csv_options(self):
        return dict(
            sep="\t",
            comment="#",
            # parse_dates=["report_timestamp"],
            skip_blank_lines=True,
            skipinitialspace=True,
            compression="zip",
        )


source = MARSRetriever



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

"""
ipython is not None when running a notebook
"""

import logging
import sys

ipython_active = None
try:
    from IPython import get_ipython

    ipython_active = get_ipython()
except Exception:
    pass

"""

Run that code in a cell:

from climetlab.core.ipython import ipython_environment
ipython_environment()

"""


def guess_which_ipython():
    if ipython_active is None:
        return (None, None)

    if ipython_active.ipython_dir == "/deepnote-config/ipython":
        return ("deepnote", None)

    if ipython_active.ipython_dir == "/home/jovyan/.ipython":
        return ("jupyter-lab", None)

    if "google.colab" in repr(ipython_active.inspector):
        return ("colab", None)

    if "IPython.terminal" in repr(ipython_active.parent):
        return ("ipython", None)

    return ("unknown", None)


def tidy(x):
    if x is None:
        return x

    if isinstance(x, (list, tuple)):
        return [tidy(y) for y in x]

    if isinstance(x, dict):
        r = {}
        for k, v in x.items():
            r[str(k)] = tidy(v)
        return r

    if isinstance(x, (int, float, str)):
        return x

    import re

    return re.sub(r" object at x\w+", repr(x), "")


def ipython_environment():
    import json

    import IPython

    r = {}
    k = IPython.get_ipython()
    for n in dir(k):
        if not callable(getattr(k, n)) and not n.startswith("__"):
            r[n] = getattr(k, n)
    print(json.dumps(tidy(r), sort_keys=True, indent=4))


def enable_ipython_login(level=logging.INFO):
    class Filter(logging.Filter):
        def filter(self, message):
            return message.levelno < logging.WARNING

    logger = logging.getLogger()
    logger.setLevel(logging.NOTSET)

    stdout = logging.StreamHandler(sys.stdout)
    stdout.setLevel(level)
    stdout.addFilter(Filter())
    logger.addHandler(stdout)

    stderr = logging.StreamHandler(sys.stderr)
    stderr.setLevel(logging.WARNING)
    logger.addHandler(stderr)


def _identity(x, **kwargs):
    return x


if ipython_active:
    from IPython.display import HTML, SVG, Image, Markdown, display

    # enable_ipython_login()
else:
    Image = _identity
    SVG = _identity
    HTML = _identity
    Markdown = _identity
    display = _identity

__all__ = ["display", "Image", "SVG", "HTML", "Markdown"]



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
import datetime


class IntHelper:
    def __init__(self, data):
        self.data = data

    def to_datetime(self):
        if self.data <= 0:
            date = datetime.datetime.utcnow() + datetime.timedelta(days=self.data)
            return datetime.datetime(date.year, date.month, date.day)
        else:
            return datetime.datetime(
                self.data // 10000, self.data % 10000 // 100, self.data % 100
            )

    def to_datetime_list(self):
        return [self.to_datetime()]


helper = IntHelper



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import pandas as pd

from climetlab.utils import download_and_cache

from . import Meteonet


class MeteonetGroundStations(Meteonet):
    """
    See https://github.com/meteofrance/meteonet
    """

    def __init__(self, domain="NW", date="20160101"):

        url = "{url}/ground_stations/{domain}_{date}.csv".format(
            url=self.URL, domain=domain, date=date
        )

        self.path = download_and_cache(url)
        self._pandas = pd.read_csv(
            self.path, parse_dates=[4], infer_datetime_format=True
        )

    def to_pandas(self):
        return self._pandas

    def plot_map(self, driver):

        north, east = self._pandas[["lat", "lon"]].max()
        south, west = self._pandas[["lat", "lon"]].min()

        driver.bounding_box(north, west, south, east)
        driver.plot_pandas(self._pandas, "lat", "lon", "t")


dataset = MeteonetGroundStations



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import hashlib
import json
import os
import sqlite3
import tempfile

from climetlab.decorators import locked
from climetlab.utils import bytes_to_string
from climetlab.utils.html import css

from .settings import SETTINGS

_connection = None


@locked
def connection():
    global _connection
    if _connection is None:
        cache_dir = SETTINGS.get("cache-directory")
        if not os.path.exists(cache_dir):
            os.makedirs(cache_dir, exist_ok=True)
        cache_db = os.path.join(cache_dir, "cache.db")
        _connection = sqlite3.connect(cache_db)
        # So we can use rows as dictionaries
        _connection.row_factory = sqlite3.Row

        _connection.execute(
            """
            CREATE TABLE IF NOT EXISTS cache (
                    path          TEXT PRIMARY KEY,
                    owner         TEXT NOT NULL,
                    args          TEXT NOT NULL,
                    creation_date TEXT NOT NULL,
                    flags         INTEGER DEFAULT 0,
                    remote_date   TEXT, -- TODO expire URLs
                    remote_tag    TEXT, -- TODO expire URLs
                    last_access   TEXT,
                    type          TEXT,
                    parent        TEXT,
                    extra         TEXT,
                    expires       INTEGER,
                    accesses      INTEGER,
                    size          INTEGER);"""
        )

    return _connection


@locked
def settings_changed():
    global _connection
    if _connection is not None:
        _connection.close()
    _connection = None


SETTINGS.on_change(settings_changed)


@locked
def update_cache():

    with connection() as db:
        update = []
        for n in db.execute("SELECT path FROM cache WHERE size IS NULL"):
            try:
                path = n[0]
                if os.path.isdir(path):
                    kind = "directory"
                    size = 0
                    for root, _, files in os.walk(path):
                        for f in files:
                            size += os.path.getsize(os.path.join(root, f))
                else:
                    kind = "file"
                    size = os.path.getsize(path)
                update.append((size, kind, path))
            except Exception:
                pass

        if update:
            db.executemany("UPDATE cache SET size=?, type=? WHERE path=?", update)
            db.commit()


@locked
def register_cache_file(path, owner, args):

    db = connection()

    now = datetime.datetime.utcnow()

    args = json.dumps(args, indent=4)

    db.execute(
        """
        UPDATE cache
        SET accesses    = accesses + 1,
            last_access = ?
        WHERE path=?""",
        (now, path),
    )

    changes = db.execute("SELECT changes()").fetchone()[0]

    if not changes:

        db.execute(
            """
            INSERT INTO cache(
                            path,
                            owner,
                            args,
                            creation_date,
                            last_access,
                            accesses)
            VALUES(?,?,?,?,?,?)""",
            (path, owner, args, now, now, 1),
        )

    db.commit()
    return not changes


def update(m, x):
    if isinstance(x, (list, tuple)):
        for y in x:
            update(m, y)
        return

    if isinstance(x, dict):
        for k, v in sorted(x.items()):
            update(m, k)
            update(m, v)
        return

    m.update(str(x).encode("utf-8"))


@locked
def cache_file(owner: str, *args, extension: str = ".cache"):

    m = hashlib.sha256()
    update(m, owner)
    update(m, args)
    path = "%s/%s-%s%s" % (
        SETTINGS.get("cache-directory"),
        owner.lower(),
        m.hexdigest(),
        extension,
    )

    if register_cache_file(path, owner, args):
        update_cache()

    return path


class TmpFile:
    def __init__(self, path):
        self.path = path

    def __del__(self):
        os.unlink(self.path)


@locked
def temp_file(extension=".tmp"):
    fd, path = tempfile.mkstemp(suffix=extension)
    os.close(fd)
    return TmpFile(path)


class Cache:
    def _repr_html_(self):

        update_cache()

        html = [css("table")]
        with connection() as db:
            for n in db.execute("SELECT * FROM cache"):
                html.append("<table class='climetlab'>")
                html.append("<td><td colspan='2'>%s</td></tr>" % (n["path"],))

                for k in [x for x in n.keys() if x != "path"]:
                    if k == "size":
                        v = bytes_to_string(n[k])
                    else:
                        v = n[k]
                    html.append("<td><td>%s</td><td>%s</td></tr>" % (k, v))
                html.append("</table>")
                html.append("<br>")
        return "".join(html)


CACHE = Cache()



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import inspect
import threading

from climetlab.normalisers import NORMALISERS


class parameters:
    def __init__(self, **kwargs):
        self.types = dict()
        for k, v in kwargs.items():
            if isinstance(v, str):
                v = v.split(":")
            if hasattr(v[0], "normalise"):
                assert len(v) == 1, v
                self.types[k] = v[0]
            else:
                self.types[k] = NORMALISERS[v[0]](*v[1:])

    def __call__(self, func):

        spec = inspect.getfullargspec(func)

        def wrapped(*args, **kwargs):

            request = dict()
            request.update(kwargs)

            for p, a in zip(spec.args, args):
                request[p] = a

            request = self.normalise(request)
            return func(**request)

        wrapped.__name__ = func.__name__

        return wrapped

    def normalise(self, request):
        result = dict(**request)

        for k, v in self.types.items():
            if k in request:
                n = v.normalise(request[k])
                if n is not None:
                    result[k] = n

        return result


def dict_args(func):
    def wrapped(*args, **kwargs):
        m = []
        p = {}
        for q in args:
            if isinstance(q, dict):
                p.update(q)
            else:
                m.append(q)
        p.update(kwargs)
        return func(*m, **p)

    wrapped.__name__ = func.__name__

    return wrapped


LOCK = threading.RLock()


def locked(func):
    def wrapped(*args, **kwargs):
        with LOCK:
            return func(*args, **kwargs)

    wrapped.__name__ = func.__name__

    return wrapped



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


class Options:
    def __init__(self, options=None):
        self._options = {} if options is None else options
        self._used_options = set()

    def __getitem__(self, name: str):
        self._used_options.add(name)
        return self._options[name]

    def __call__(self, name: str, default):
        self._used_options.add(name)
        return self._options.get(name, default)

    def provided(self, name: str) -> bool:
        return name in self._options

    def check_unused(self):
        unused = set(self._options.keys()) - self._used_options
        if unused:
            raise TypeError(
                "".join(
                    [
                        "Unused argument%s: " % ("s" if len(unused) > 1 else "",),
                        ", ".join("%s=%s" % (x, self._options[x]) for x in unused),
                    ]
                )
            )

    def update_if_not_set(self, **kwargs):
        for k, v in kwargs.items():
            if k not in self._options:
                self._options[k] = v



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#
from climetlab.helpers import helper


class BoundingBox:
    def __init__(self, *, north, west, south, east):
        # Convert to float as these values may come from Numpy
        self.north = min(float(north), 90.0)
        self.west = float(west)
        self.south = max(float(south), -90.0)
        self.east = float(east)

        if self.north <= self.south:
            raise ValueError(
                "North (%s) must be greater than south (%s)"
                % (
                    self.north,
                    self.south,
                )
            )

        if self.west == self.east:
            raise ValueError("West (%s) is equal to east (%s)" % (west, east))

        while self.east < self.west:
            self.east += 360

        while self.east - self.west > 360:
            self.east -= 360

        while self.east >= 360 and self.west >= 360:
            self.east -= 360
            self.west -= 360

        while self.east < -180 and self.west < -180:
            self.east += 360
            self.west += 360

    def __repr__(self):
        return "BoundingBox(north=%g,west=%g,south=%g,east=%g)" % (
            self.north,
            self.west,
            self.south,
            self.east,
        )

    def __eq__(self, other):
        if self.__class__ is not other.__class__:
            return False
        return self.as_tuple() == other.as_tuple()

    @property
    def width(self):
        return self.east - self.west

    @property
    def height(self):
        return self.north - self.south

    def merge(self, other):

        west1, east1 = self.west, self.east
        west2, east2 = other.west, other.east

        while west1 < 0 or east1 < 0:
            west1 += 360
            east1 += 360

        while west2 < 0 or east2 < 0:
            west2 += 360
            east2 += 360

        if abs(west1 - (west2 + 360)) < abs(west1 - west2):
            east2, west2 = east2 + 360, west2 + 360
        elif abs(west2 - (west1 + 360)) < abs(west2 - west1):
            east1, west1 = east1 + 360, west1 + 360

        # print(west1, east1, west2, east2)
        return BoundingBox(
            north=max(self.north, other.north),
            west=min(west1, west2),
            south=min(self.south, other.south),
            east=max(east1, east2),
        )

    def add_margins(self, margins):

        if isinstance(margins, str) and margins[-1] == "%":
            margins = int(margins[:-1]) / 100.0
            margins = max(
                (self.north - self.south) * margins, (self.east - self.west) * margins
            )

        # TODO:check east/west
        margins_lat = margins
        margins_lon = margins

        if self.east - self.west > 360 - 2 * margins:
            margins = (360 - (self.east - self.west)) / 2.0

        return BoundingBox(
            north=self.north + margins_lat,
            west=self.west - margins_lon,
            south=self.south - margins_lat,
            east=self.east + margins_lon,
        )

    def as_list(self):
        return [self.north, self.west, self.south, self.east]

    def as_tuple(self):
        return (self.north, self.west, self.south, self.east)

    def as_dict(self):
        return dict(north=self.north, west=self.west, south=self.south, east=self.east)


def to_bounding_box(obj):

    if isinstance(obj, BoundingBox):
        return obj

    if isinstance(obj, (list, tuple)):
        return BoundingBox(north=obj[0], west=obj[1], south=obj[2], east=obj[3])

    if getattr(obj, "to_bounding_box", None) is None:
        obj = helper(obj)

    return to_bounding_box(obj.to_bounding_box())



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os

import pytest
import yaml

import climetlab.plotting
from climetlab import settings
from climetlab.core.data import clear_cache, get_data_entry
from climetlab.core.plugins import directories


def check_user_defined_objects(collection, setting, obj, tree, get_list, get_entry):

    # Clear cache
    clear_cache()

    paths = settings.get(setting)
    assert isinstance(paths, (list, tuple)), paths
    assert len(paths) > 0

    for i, path in enumerate(paths):

        name = "pytest-%s-%s" % (tree[1], i)

        if os.path.exists(path) and os.path.isdir(path):
            assert path in directories(), directories()

        if not os.path.exists(path):
            os.mkdir(path)

        with open(os.path.join(path, "%s.yaml" % (name,)), "w") as f:
            a = obj
            for t in tree[:-1]:
                a = a[t]
            a[tree[-1]] = i
            yaml.dump(obj, f, default_flow_style=False)

    for i in range(len(paths)):
        name = "pytest-%s-%s" % (tree[1], i)
        get_data_entry(collection, name)
        assert name in get_list()
        p = get_entry(name).data

        a = p
        for t in tree:
            a = a[t]

        assert a == i

    # TODO: Move to tear-down
    for i, path in enumerate(paths):
        name = "pytest-%s-%s" % (tree[1], i)
        os.unlink(os.path.join(path, "%s.yaml" % (name,)))


def test_user_projections():
    check_user_defined_objects(
        "projections",
        "projections-directories",
        {"magics": {"mmap": {}}},
        ["magics", "mmap", "subpage_lower_left_latitude"],
        climetlab.plotting.projections,
        climetlab.plotting.projection,
    )


def test_user_styles_msymb():
    check_user_defined_objects(
        "styles",
        "styles-directories",
        {"magics": {"msymb": {}}},
        ["magics", "msymb", "symbol_marker_index"],
        climetlab.plotting.styles,
        climetlab.plotting.style,
    )


def test_user_styles_mcont():
    check_user_defined_objects(
        "styles",
        "styles-directories",
        {"magics": {"mcont": {}}},
        ["magics", "mcont", "contour_line_thickness"],
        climetlab.plotting.styles,
        climetlab.plotting.style,
    )


def test_user_layers():
    check_user_defined_objects(
        "layers",
        "layers-directories",
        {"magics": {"mcoast": {}}},
        ["magics", "mcoast", "map_grid_frame_thickness"],
        climetlab.plotting.layers,
        climetlab.plotting.layer,
    )


def test_settings():

    settings.reset()

    assert settings.get("plotting-options") == {}, "Check 1"
    settings.set("plotting-options", width=400)
    assert settings.get("plotting-options") == {"width": 400}
    settings.reset("plotting-options")
    assert settings.get("plotting-options") == {}, "Check 2"
    settings.set("plotting-options", {"width": 400})
    assert settings.get("plotting-options") == {"width": 400}
    settings.reset()
    assert settings.get("plotting-options") == {}, "Check 3"

    with pytest.raises(TypeError):
        settings.set("plotting-options", 3)

    settings.set("styles-directories", ["/a", "/b"])
    assert settings.get("styles-directories") == ["/a", "/b"]

    settings.set("styles-directories", "/c", "/d")
    assert settings.get("styles-directories") == ["/c", "/d"]

    with pytest.raises(KeyError):
        settings.set("test", 42)

    with pytest.raises(KeyError):
        settings.get("test")

    settings.reset()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

# TODO: Use magics types

from . import magics_keys_parameters
from .colour import Colour


class NoConvertion:
    def convert(self, name, value):
        return value


class Bool:
    def convert(self, name, value):
        c = {"on": True, "yes": True, "off": False, "no": False, None: False}
        return bool(c.get(value, value))


class Float:
    def convert(self, name, value):
        return float(value)


class Int:
    def convert(self, name, value):
        return int(value)


class String:
    def convert(self, name, value):
        c = {None: "", False: "off", True: "on"}
        return str(c.get(value, value))


class ColourList:
    def convert(self, name, value):
        assert isinstance(value, (list, tuple))
        c = Colour()
        return [c.convert(name, x) for x in value]


class FloatList:
    def convert(self, name, value):
        assert isinstance(value, (list, tuple))
        c = Float()
        return [c.convert(name, x) for x in value]


class IntList:
    def convert(self, name, value):
        assert isinstance(value, (list, tuple))
        c = Int()
        return [c.convert(name, x) for x in value]


class StringList:
    def convert(self, name, value):
        assert isinstance(value, (list, tuple))
        c = String()
        return [c.convert(name, x) for x in value]


def convert(action, args):

    magics_keys = magics_keys_parameters(action)

    converted = {}
    for k, v in args.items():
        klass = magics_keys.get(k, {}).get("type", "NoConvertion")
        klass = globals().get(klass, NoConvertion)()

        converted[k] = klass.convert(k, v)

    return converted



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab.core import docstring
from climetlab.core.data import data_entries, get_data_entry
from climetlab.core.ipython import display
from climetlab.core.settings import SETTINGS
from climetlab.helpers import helper

from .drivers.magics.driver import Driver
from .options import Options

OPTIONS = {}


def options(**kwargs):
    global OPTIONS
    OPTIONS = kwargs


def projection(name):
    return get_data_entry("projections", name)


def layer(name):
    return get_data_entry("layers", name)


def style(name):
    return get_data_entry("styles", name)


def projections():
    return sorted(e.name for e in data_entries("projections"))


def layers():
    return sorted(e.name for e in data_entries("layers"))


def styles():
    return sorted(e.name for e in data_entries("styles"))


class Plot:
    """[summary]"""

    def __init__(self, kwargs):
        options = {}
        options.update(SETTINGS.get("plotting-options", {}))
        options.update(OPTIONS)
        options.update(kwargs)
        self.driver = Driver(Options(options))

    def plot_map(self, data=None, **kwargs):

        # try:
        #     iter(data)
        #     data = list(data)
        # except Exception:
        #     pass

        if not isinstance(data, (list, tuple)):
            data = [data]

        for d in data:
            if getattr(d, "plot_map", None) is None:
                d = helper(d)

            d.plot_map(self.driver)

        options = Options(kwargs)
        self.driver.apply_options(options)
        options.check_unused()

        return self

    def wms_layers(self):
        return self.driver.wms_layers()

    def show(self):
        return display(self.driver.show())

    def macro(self) -> list:
        return self.driver.macro()

    def save(self, path):
        return self.driver.save(path)


def new_plot(**kwargs) -> Plot:
    """[summary]

    :return: [description]
    :rtype: Plot
    """
    return Plot(kwargs)


@docstring()
def plot_map(data=None, **kwargs):
    """Plot any data on a map.

    Args:
        data ([any]): [description]
    """

    p = new_plot(**kwargs)
    p.plot_map(data)
    p.show()


Plot.plot_map.__doc__ = plot_map.__doc__


def interactive_map(data=None, **kwargs):
    from climetlab.plotting.wms import interactive_map as wms_map

    return wms_map(data, **kwargs)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys

import pytest

import climetlab as cml
from climetlab import dataset, load_dataset


def test_dataset_1():
    load_dataset("sample-bufr-data")


@pytest.mark.skipif(sys.version_info < (3, 7), reason="Version 3.7 or greater needed")
def test_dataset_2():
    dataset.sample_bufr_data()


def test_era5_temperature():
    if not os.path.exists(os.path.expanduser("~/.cdsapirc")):
        pytest.skip("No ~/.cdsapirc")

    ds = cml.load_dataset(
        "era5-temperature", period=(1979, 1982), domain="France", time=12
    )


def test_datetime():
    if not os.path.exists(os.path.expanduser("~/.cdsapirc")):
        pytest.skip("No ~/.cdsapirc")

    data = cml.load_dataset(
        "era5-temperature", domain="france", period=(1980,), time=12
    )
    data["1980-12-09 12:00"]
    with pytest.raises(NotImplementedError):
        data.sel(date="1980-12-09 12:00")


def test_pandas_filter():
    data = cml.load_dataset("hurricane-database", bassin="atlantic")
    irma = data.to_pandas(name="irma", year=2017)
    assert len(irma) == 66


if __name__ == "__main__":
    test_datetime()

if __name__ == "__main__":
    test_dataset_1()



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

# Keep linters happy
# N801 = classes should start with uppercase


class docstring:  # noqa: N801
    def __init__(self):
        pass

    def __call__(self, func):
        # func.__doc__ += "Decorated...."
        return func



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

# See https://www.aoml.noaa.gov/hrd/hurdat/Data_Storm.html

import functools

import numpy as np
import pandas as pd

from climetlab.utils import download_and_cache
from climetlab.utils.dates import parse_date

from . import Dataset

SIGN = {"N": 1, "W": -1, "E": 1, "S": -1}

TYPE = {
    "C": "Closest approach to a coast, not followed by a landfall",
    "G": "Genesis",
    "I": "An intensity peak in terms of both pressure and wind",
    "L": "Landfall (center of system crossing a coastline)",
    "P": "Minimum in central pressure",
    "R": "Provides additional detail on the intensity of the cyclone when rapid changes are underway",
    "S": "Change of status of the system",
    "T": "Provides additional detail on the track (position) of the cyclone",
    "W": "Maximum sustained wind speed",
}

STATUS = {
    "TD": "Tropical cyclone of tropical depression intensity (< 34 knots)",
    "TS": "Tropical cyclone of tropical storm intensity (34-63 knots)",
    "HU": "Tropical cyclone of hurricane intensity (> 64 knots)",
    "EX": "Extratropical cyclone (of any intensity)",
    "SD": "Subtropical cyclone of subtropical depression intensity (< 34 knots)",
    "SS": "Subtropical cyclone of subtropical storm intensity (> 34 knots)",
    "LO": "A low that is neither a tropical cyclone, a subtropical cyclone, nor an extratropical cyclone (of any intensity)",
    "WV": "Tropical Wave (of any intensity)",
    "DB": "Disturbance (of any intensity)",
}


# https://en.wikipedia.org/wiki/Saffir–Simpson_scale
def category(knots):

    if knots < 83:
        return 1

    if knots < 96:
        return 2

    if knots < 113:
        return 3

    if knots < 137:
        return 4

    return 5


URLS = {
    "atlantic": "https://www.aoml.noaa.gov/hrd/hurdat/hurdat2.txt",
    "pacific": "https://www.aoml.noaa.gov/hrd/hurdat/hurdat2-nepac.html",
}


class HurricaneDatabase(Dataset):

    home_page = "https://www.aoml.noaa.gov/hrd/hurdat/Data_Storm.html"

    def __init__(self, *, bassin="atlantic", url=None):

        if url is None:
            url = URLS[bassin.lower()]

        path = download_and_cache(url)
        p = []
        with open(path) as f:
            lines = f
            for line in lines:
                if line[0] in (" ", "<", "\n"):
                    continue

                bassin = line[0:2]
                number = int(line[2:4])
                year = int(line[4:8])
                name = line[18:28].strip().lower()
                # id = line[0:8]

                # http://www.aoml.noaa.gov/hrd/hurdat/hurdat2-format-may2015.pdf

                for _ in range(0, int(line[33:36])):
                    line = next(lines)
                    knots = float(line[38:41])
                    pressure = np.NaN if line[43] == "-" else float(line[43:47])
                    time = "%s-%s-%sZ%s:%s" % (
                        line[0:4],
                        line[4:6],
                        line[6:8],
                        line[10:12],
                        line[12:14],
                    )
                    p.append(
                        dict(
                            # id=id,
                            bassin=bassin,
                            number=number,
                            year=year,
                            name=name,
                            time=parse_date(time),
                            type=line[16],
                            status=line[19:21],
                            lat=float(line[23:27]) * SIGN[line[27]],
                            lon=float(line[30:35]) * SIGN[line[35]],
                            knots=knots,
                            category=category(knots),
                            pressure=pressure,
                        )
                    )

        self.cyclones = self.annotate(pd.DataFrame(p), style="cyclone-track")

    def to_pandas(self, **kwargs):
        if kwargs:
            df = self.cyclones
            f = [df[k] == v for k, v in kwargs.items()]
            return df[functools.reduce(lambda a, b: a & b, f)]
        return self.cyclones


dataset = HurricaneDatabase



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import logging
import os
from importlib import import_module

import entrypoints

import climetlab

from .settings import SETTINGS

LOG = logging.getLogger(__name__)

CACHE = {}

PLUGINS = {}


def _load_plugins(kind):
    plugins = {}
    for e in entrypoints.get_group_all(f"climetlab.{kind}s"):
        plugins[e.name.replace("_", "-")] = e
    return plugins


def load_plugins(kind):
    if PLUGINS.get(kind) is None:
        PLUGINS[kind] = _load_plugins(kind)
    return PLUGINS[kind]


def find_plugin(directory, name, loader):

    plugins = load_plugins(loader.kind)

    if name in plugins:
        plugin = plugins[name]
        return loader.load_entry(plugin)

    n = len(directory)
    for path, _, files in os.walk(directory):
        path = path[n:]
        for f in files:
            if f.endswith(".yaml") and f[:-5] == name:
                full = os.path.join(directory, path, f)
                return loader.load_yaml(full)

            if f.endswith(".py"):
                p = os.path.join(path, f[:-3])
                if p[0] != "/":
                    p = "/" + p
                if p[1:].replace("/", "-").replace("_", "-") == name:
                    return loader.load_module(p.replace("/", "."))

    plugins_str = ",".join(["'" + str(p) + "'" for p in plugins.keys()])
    raise Exception(
        "Cannot find %s '%s' (valid values are %s)" % (loader.kind, name, plugins_str)
    )


def directories():

    result = []
    for conf in ("styles-directories", "projections-directories", "layers-directories"):
        for d in SETTINGS.get(conf):
            if os.path.exists(d) and os.path.isdir(d):
                result.append(d)

    for kind in ("dataset", "source"):
        for _, v in load_plugins(kind).items():
            try:
                module = import_module(v.module_name)
                result.append(os.path.dirname(module.__file__))
            except Exception:
                LOG.error("Cannot load module %s", v.module_name, exc_info=True)

    result.append(os.path.dirname(climetlab.__file__))

    return result



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import numpy as np
import xarray as xr

from climetlab.utils import download_and_cache

from . import Meteonet

"""
rainfall_diff_quality-code
rainfall_mean_quality-code
reflectivity_new
reflectivity_old
"""


class MeteonetRadar(Meteonet):
    """
    See https://github.com/meteofrance/meteonet
    """

    def __init__(self, domain="NW", variable="rainfall", year=2016, month=8, part=3):

        url = "{url}/radar/radar_coords_{domain}.npz".format(
            url=self.URL, domain=domain
        )

        coords = np.load(download_and_cache(url), allow_pickle=True)

        resolution = 0.01

        lats = coords["lats"] - resolution / 2
        lons = coords["lons"] + resolution / 2

        url = "{url}/radar/{variable}_{domain}_{year}_{month:02d}.{part}.npz".format(
            url=self.URL,
            domain=domain,
            variable=variable,
            year=year,
            month=month,
            part=part,
        )

        path = download_and_cache(url)
        content = np.load(path, allow_pickle=True)
        data = content["data"]
        times = content["dates"]
        # missing = content['miss_dates']

        self.variable = variable

        ds = xr.Dataset(
            {
                variable: (["time", "y", "x"], data),
                "x": (["x"], range(0, data.shape[2])),
                "y": (["y"], range(0, data.shape[1])),
            },
            coords={
                "lon": (["y", "x"], lons),
                "lat": (["y", "x"], lats),
                "time": times,
            },
        )

        self.north = np.amax(lats)
        self.south = np.amin(lats)
        self.east = np.amax(lons)
        self.west = np.amin(lons)

        ds["lon"].attrs["standard_name"] = "longitude"
        ds["lat"].attrs["standard_name"] = "latitude"
        ds["time"].attrs["standard_name"] = "time"
        ds["x"].attrs["axis"] = "X"
        ds["y"].attrs["axis"] = "Y"

        self._xarray = ds

    def to_xarray(self):
        return self._xarray

    def plot_map(self, driver):
        driver.bounding_box(self.north, self.west, self.south, self.east)

        dimensions = {"time": 0}

        driver.plot_xarray(self._xarray, self.variable, dimensions)
        driver.style("meteonet-samples-radar-{}".format(self.variable))


dataset = MeteonetRadar



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import re


def urlify(text):
    return re.sub(r"(https?://.*\S)", r'<a href="\1" target="_blank">\1</a>', text)


def css(name):
    path = os.path.join(os.path.dirname(os.path.dirname(__file__)), "data", "css", name)
    with open(path + ".css") as f:
        return "<style>%s</style>" % (f.read(),)


def table(obj):

    style = css("table")

    table = """
<h4>{name}</h4>
<table class="climetlab">
<tr><td><b>Home page</b></td><td>{home_page}</td></tr>
<tr><td><b>Documentation</b></td><td>{documentation}</td></tr>
<tr><td><b>Citation</b></td><td><pre>{citation}</pre></td></tr>
<tr><td><b>Licence</b></td><td>{licence}</td></tr>
</table>
        """.format(
        name=obj.name,
        home_page=urlify(obj.home_page),
        licence=urlify(obj.licence),
        citation=obj.citation,
        documentation=urlify(obj.documentation),
    )

    return style + table



# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import datetime
import logging
import weakref

import eccodes

from climetlab.decorators import dict_args
from climetlab.utils.bbox import BoundingBox

from . import Reader

LOG = logging.getLogger(__name__)

# return eccodes.codes_new_from_file(self.file, eccodes.CODES_PRODUCT_GRIB)

# See https://pymotw.com/2/weakref/


class CodesHandle:
    def __init__(self, handle, path, offset):
        self.handle = handle
        self.path = path
        self.offset = offset

    def __del__(self):
        eccodes.codes_release(self.handle)

    def get(self, name):
        try:
            if name == "values":
                return eccodes.codes_get_values(self.handle)
            if name in ("distinctLatitudes", "distinctLongitudes"):
                return eccodes.codes_get_double_array(self.handle, name)
            return eccodes.codes_get(self.handle, name)
        except eccodes.KeyValueNotFoundError:
            return None


class CodesReader:
    def __init__(self, path):
        self.path = path
        self.file = open(self.path, "rb")

    def __del__(self):
        try:
            self.file.close()
        except Exception:
            pass

    def at_offset(self, offset):
        self.file.seek(offset, 0)
        return self.__next__()

    def __iter__(self):
        return self

    def __next__(self):
        offset = self.file.tell()
        handle = eccodes.codes_new_from_file(self.file, eccodes.CODES_PRODUCT_GRIB)
        if not handle:
            raise StopIteration()
        return CodesHandle(handle, self.path, offset)

    @property
    def offset(self):
        return self.file.tell()


def cb(r):
    print("Delete", r)


class GribField:
    def __init__(self, *, handle=None, reader=None, offset=None):
        self._handle = handle
        self._reader = reader
        self._offset = offset

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        pass

    @property
    def path(self):
        return self.handle.path

    @property
    def handle(self):
        if self._handle is None:
            assert self._offset is not None
            assert self._reader is not None
            self._handle = self._reader.at_offset(self._offset)
        return self._handle

    @property
    def values(self):
        return self.handle.get("values")

    @property
    def offset(self):
        if self._offset is None:
            self._offset = int(self.handle.get("offset"))
        return self._offset

    @property
    def shape(self):
        return self.handle.get("Nj"), self.handle.get("Ni")

    def plot_map(self, driver):
        driver.bounding_box(
            north=self.handle.get("latitudeOfFirstGridPointInDegrees"),
            south=self.handle.get("latitudeOfLastGridPointInDegrees"),
            west=self.handle.get("longitudeOfFirstGridPointInDegrees"),
            east=self.handle.get("longitudeOfLastGridPointInDegrees"),
        )
        driver.plot_grib(self.path, self.handle.get("offset"))

    def to_numpy(self):
        return self.values.reshape(self.shape)

    def __repr__(self):
        return "GribField(%s,%s,%s,%s,%s,%s)" % (
            self.handle.get("shortName"),
            self.handle.get("levelist"),
            self.handle.get("date"),
            self.handle.get("time"),
            self.handle.get("step"),
            self.handle.get("number"),
        )

    def _grid_definition(self):
        return dict(
            north=self.handle.get("latitudeOfFirstGridPointInDegrees"),
            south=self.handle.get("latitudeOfLastGridPointInDegrees"),
            west=self.handle.get("longitudeOfFirstGridPointInDegrees"),
            east=self.handle.get("longitudeOfLastGridPointInDegrees"),
            south_north_increment=self.handle.get("jDirectionIncrementInDegrees"),
            west_east_increment=self.handle.get("iDirectionIncrementInDegrees"),
        )

    def field_metadata(self):
        m = self._grid_definition()
        for n in ("shortName", "units", "paramId"):
            p = self.handle.get(n)
            if p is not None:
                m[n] = str(p)
        m["shape"] = self.shape
        return m

    def helper(self):
        return self

    def datetime(self):
        date = self.handle.get("date")
        time = self.handle.get("time")
        return datetime.datetime(
            date // 10000, date % 10000 // 100, date % 100, time // 100, time % 100
        )

    def valid_datetime(self):
        step = self.handle.get("endStep")
        return self.datetime() + datetime.timedelta(hours=step)

    def to_datetime_list(self):
        return [self.valid_datetime()]

    def to_bounding_box(self):
        return BoundingBox(
            north=self.handle.get("latitudeOfFirstGridPointInDegrees"),
            south=self.handle.get("latitudeOfLastGridPointInDegrees"),
            west=self.handle.get("longitudeOfFirstGridPointInDegrees"),
            east=self.handle.get("longitudeOfLastGridPointInDegrees"),
        )

    def _attributes(self, names):
        result = {}
        for name in names:
            result[name] = self.handle.get(name)
        return result


class GRIBIterator:
    def __init__(self, path):
        self.path = path
        self.reader = CodesReader(path)

    def __repr__(self):
        return "GRIBIterator(%s)" % (self.path,)

    def __next__(self):
        offset = self.reader.offset
        handle = next(self.reader)
        return GribField(handle=handle, reader=self.reader, offset=offset)

    def __iter__(self):
        return self


class GRIBFilter:
    def __init__(self, reader, filter):
        self._reader = reader
        self._filter = dict(**filter)

    def __repr__(self):
        return "GRIBFilter(%s, %s)" % (self._reader, self._filter)

    def __iter__(self):
        return GRIBIterator(self.path)


class GRIBMultiFileReader(Reader):
    def __init__(self, source, readers):
        super().__init__(source, "/-multi-")


class OLDGRIBReader(Reader):
    def __init__(self, source, path):
        super().__init__(source, path)
        self._fields = None
        self._reader = None

    def __repr__(self):
        return "GRIBReader(%s)" % (self.path,)

    def __iter__(self):
        return GRIBIterator(self.path)

    def _items(self):
        if self._fields is None:
            self._fields = []
            for f in self:
                self._fields.append(f.offset)
        return self._fields

    def __getitem__(self, n):
        if self._reader is None:
            self._reader = CodesReader(self.path)
        return GribField(reader=self._reader, offset=self._items()[n])

    def __len__(self):
        return len(self._items())

    def to_xarray(self):
        import xarray as xr

        params = self.source.cfgrib_options()
        ds = xr.open_dataset(self.path, engine="cfgrib", **params)
        return self.source.post_xarray_open_dataset_hook(ds)

    @dict_args
    def sel(self, **kwargs):
        return GRIBFilter(self, kwargs)

    def multi_merge(source, readers):
        return GRIBMultiFileReader(source, readers)


class GRIBReader(Reader):
    def __init__(self, source, paths=None, fields=[], filter=None, unfiltetered=True):
        super().__init__(source, paths)
        self._fields = [f for f in fields]
        self._unfiltetered = unfiltetered

        if paths is not None:
            if not isinstance(paths, (list, tuple)):
                paths = [paths]
            for path in paths:
                for f in GRIBIterator(path):
                    self._fields.append(f)

    def __getitem__(self, n):
        return self._fields[n]

    def __len__(self):
        return len(self._fields)

    def _attributes(self, names):
        result = []
        for field in self:
            result.append(field._attributes(names))
        return result

    def multi_merge(source, readers):
        fields = []
        paths = []
        for r in readers:
            fields += r._fields
            if isinstance(r.path, (list, tuple)):
                paths += list(r.path)
            else:
                paths.append(r.path)

        return GRIBReader(
            source,
            paths=paths,
            # fields=fields,
            unfiltetered=all(r._unfiltetered for r in readers),
        )

    def to_xarray(self):
        assert self._unfiltetered
        assert self.path

        import xarray as xr

        params = self.source.cfgrib_options()
        if isinstance(self.path, (list, tuple)):
            ds = xr.open_mfdataset(self.path, engine="cfgrib", **params)
        else:
            ds = xr.open_dataset(self.path, engine="cfgrib", **params)
        return self.source.post_xarray_open_dataset_hook(ds)


class Field:
    def __init__(self, field):
        self.field = field

    def __getitem__(self, key):
        print("field", key)
        v = self.field.handle.get(key)
        return v


class FieldSetIndex:
    def __init__(
        self,
        source,
        *,
        filter_by_keys={},
        grib_errors="warn",
        index_keys=[],
        read_keys=[],
        items=None,
    ):
        self.source = source
        self.filter_by_keys = filter_by_keys
        self.grib_errors = grib_errors
        self.index_keys = index_keys
        self.read_keys = read_keys
        self.items = items

    def __call__(self, *, grib_errors, index_keys, read_keys, **kwargs):
        return FieldSetIndex(
            source=self.source,
            filter_by_keys=self.filter_by_keys,
            grib_errors=grib_errors,
            index_keys=index_keys,
            read_keys=read_keys,
            items=self.items,
        )

    def subindex(self, filter_by_keys={}, **kwargs):
        query = dict(**self.filter_by_keys)
        query.update(filter_by_keys)
        query.update(kwargs)
        return FieldSetIndex(
            source=self.source,
            filter_by_keys=query,
            grib_errors=self.grib_errors,
            index_keys=self.index_keys,
            read_keys=self.read_keys,
            items=self.items,
        )

    def _valid(self, item):
        for k, v in self.filter_by_keys.items():
            if item.get(k) != v:
                return False
        return True

    def _valid_items(self):
        if self.items is None:
            self.items = self.source._attributes(self.index_keys)

        return [i for i in self.items if self._valid(i)]

    def __getitem__(self, key):

        x = [i[key] for i in self._valid_items()]
        if None in x:
            return []

        print("GET", key, x)
        return list(set(x))

    def getone(self, key):
        return self[key][0]

    @property
    def offsets(self):
        for i, item in enumerate(self._valid_items()):
            yield item, i

    @property
    def filestream(self):
        return self

    @property
    def path(self):
        return None

    def first(self):
        return Field(self.source[0])



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab.utils import consume_args


def consume(func1, func2, *args, **kwargs):
    args_1, kwargs_1, args_2, kwargs_2 = consume_args(func1, *args, **kwargs)

    r1 = func1(*args_1, **kwargs_1)
    r2 = func2(*args_2, **kwargs_2)

    return r1, r2


def f_void():
    return 0


def f_x(x):
    assert x == 1
    return -x


def f_y(y):
    assert y == 2
    return 2 * y


def g_x(x=1, *, a=9):
    assert x == 1
    return -x + a


def g_y(y=2, *, b=9):
    assert y == 2
    return 2 * y + b


def test_consume():
    assert consume(f_void, f_void) == (0, 0)
    assert consume(f_void, f_x, 1) == (0, -1)
    assert consume(f_x, f_void, 1) == (-1, 0)
    assert consume(f_void, f_x, x=1) == (0, -1)
    assert consume(f_x, f_void, x=1) == (-1, 0)
    assert consume(f_x, f_y, 1, 2) == (-1, 4)
    assert consume(f_y, f_x, 2, 1) == (4, -1)
    assert consume(f_x, f_y, 1, y=2) == (-1, 4)
    assert consume(f_x, f_y, x=1, y=2) == (-1, 4)
    assert consume(f_y, f_x, x=1, y=2) == (4, -1)
    assert consume(g_y, g_x, 2, 1) == (13, 8)
    assert consume(g_y, g_x, 2, 1, a=5) == (13, 4)
    assert consume(g_y, g_x, 2, 1, b=5) == (9, 8)
    assert consume(g_y, g_x) == (13, 8)
    assert consume(g_y, g_x, b=5) == (9, 8)
    assert consume(g_y, g_x, y=2) == (13, 8)
    assert consume(g_y, g_x, y=2, a=6) == (13, 5)
    assert consume(g_y, g_x, x=1, a=6) == (13, 5)
    assert consume(g_y, g_x, x=1, b=4) == (8, 8)
    assert consume(g_y, g_x, y=2, b=4) == (8, 8)

    # Ambiguous use may not work as expected.
    # assert consume(f_y, g_y, y=2) == (4,13)
    # assert consume(g_y, f_y, 2, y=2) == (-1, 0)


if __name__ == "__main__":
    test_consume()



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#


import io
import os.path
import sys

import setuptools


def read(fname):
    file_path = os.path.join(os.path.dirname(__file__), fname)
    return io.open(file_path, encoding="utf-8").read()


version = None
for line in read("climetlab/__init__.py").split("\n"):
    if line.startswith("__version__"):
        version = line.split("=")[-1].strip()[1:-1]


assert version

install_requires = []

if sys.version_info < (3, 7):
    install_requires += ["cython", "pandas==1.1.5"]
else:
    install_requires += ["cython", "pandas"]


install_requires += [
    # need to install  to avoid conflict between aiohttp (dependency of s3fs) and requests (cdsapi)
    # "chardet>=3.0,<4.0",
    # "aiohttp>=3.7.2",
    # --
    "requests",
    # "zarr",
    # "s3fs",
    "numpy",
    "xarray>=0.17.0",
    "netcdf4",
    "cfgrib>=0.9.8.4",
    "cdsapi",
    "ecmwf-api-client>=1.6.1",
    "tqdm",
    "requests",
    "eccodes>=0.9.9",
    "magics>=1.5.6",
    "ecmwflibs>=0.1.2",
    "pdbufr",
    "pyodc",
    "dask",
    "toolz",
    "pyyaml",
    "markdown",
    "entrypoints",
    "branca==0.3.1",  # See https://github.com/python-visualization/branca/issues/81"
    "folium>=0.12.1",
]

extras_require = {}  # {"interactive": ["skinnywms", "folium"]}


full = []
for k, v in extras_require.items():
    full += v
full += install_requires

extras_require["full"] = full


setuptools.setup(
    name="climetlab",
    version=version,
    author="ECMWF",
    author_email="software.support@ecmwf.int",
    license="Apache 2.0",
    url="https://github.com/ecmwf/climetlab",
    description="Handling of climate/meteorological data",
    long_description=read("README.md"),
    long_description_content_type="text/markdown",
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=install_requires,
    extras_require=extras_require,
    zip_safe=True,
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Operating System :: OS Independent",
    ],
    tests_require=["pytest", "nbconvert", "jupyter", "pytest-cov"],
    test_suite="tests",
)



# (C) Copyright 2021 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab.utils.dates import to_date_list


class DateListNormaliser:
    def __init__(self, format=None):
        self.format = format

    def normalise(self, dates):
        dates = to_date_list(dates)
        if self.format is not None:
            dates = [d.strftime(self.format) for d in dates]
        return dates



#!/usr/bin/env python3
# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

import os
import sys
import traceback

import climetlab
import climetlab.plotting

DOCS = os.path.join(os.path.dirname(os.path.dirname(os.path.dirname(__file__))), "docs")


def plot_projection(name, path):
    print("plot_projection", name, file=sys.stderr)
    climetlab.plot_map(None, projection=name, background="land-sea", path=path)
    print("done", name, file=sys.stderr)


def plot_layer(name, path):
    print("plot_layer", name, file=sys.stderr)
    climetlab.plot_map(None, background=False, foreground=name, path=path)
    print("done", name, file=sys.stderr)


def get_data(how, name, to_pandas=None, **kwargs):
    data = how(name, **kwargs)

    if to_pandas is not None:
        return data.to_pandas(**to_pandas)

    try:
        return data[0]
    except Exception as e:
        print("Exection", e)
        return data


def get_dataset(name, to_pandas=None, **kwargs):
    return get_data(climetlab.load_dataset, name, to_pandas, **kwargs)


def get_source(name, to_pandas=None, **kwargs):
    return get_data(climetlab.load_source, name, to_pandas, **kwargs)


def plot_style(name, path):
    print("plot_style", name, file=sys.stderr)
    yaml = climetlab.plotting.style(name).data

    data = None
    gallery = yaml.get("gallery", {})

    if "sample" in gallery:
        sample = gallery["sample"]
        if "source" in sample:
            source = sample["source"]
            data = get_source(
                source["name"], source.get("to_pandas"), **source.get("args", {})
            )
        else:
            dataset = sample["dataset"]
            data = get_dataset(
                dataset["name"], dataset.get("to_pandas"), **dataset.get("args", {})
            )
    else:
        if "msymb" in yaml["magics"]:
            data = get_dataset(
                "sample-bufr-data",
                to_pandas={"columns": ("latitude", "longitude"), "filters": {}},
            )
        if "mcont" in yaml["magics"]:
            data = get_dataset("sample-grib-data")

    extra = gallery.get("plot_map", {})

    climetlab.plot_map(data, style=name, path=path, **extra)
    print("done", name, file=sys.stderr)


def output(title, collection, plotter):

    print()
    print(title[0].upper() + title[1:])
    print("-" * len(title))
    print()

    path = os.path

    for p in collection:
        print()
        print(p)
        print("^" * len(p))

        image = "_static/gallery/%s/%s.svg" % (title, p)
        path = os.path.join(DOCS, image)

        if not os.path.exists(path):
            try:
                os.makedirs(os.path.dirname(path))
            except FileExistsError:
                pass
            try:
                plotter(p, path)
            except Exception as e:
                print(path)
                print(traceback.format_exc())

        print()
        print(".. image::", "/" + image)
        print("   :width: 600")
        print()


def execute():

    output("layers", climetlab.plotting.layers(), plot_layer)
    output("styles", climetlab.plotting.styles(), plot_style)
    output("projections", climetlab.plotting.projections(), plot_projection)


if __name__ == "__main__":
    execute()



from climetlab import new_plot
from climetlab.core.caching import temp_file

from ._folium import make_map


def interactive_map(obj, **kwargs):

    h = 100 * 3

    tmp = temp_file(".svg")
    p = new_plot(
        projection="web-mercator",
        width=6 * 1024,
        width_cm=h,
        height_cm=h,
        frame=False,
        foreground=False,
        background=False,
    )
    p.plot_map(obj)
    bbox = p.save(tmp.path)

    return make_map(tmp.path, bbox)



#!/usr/bin/env python3

# (C) Copyright 2020 ECMWF.
#
# This software is licensed under the terms of the Apache Licence Version 2.0
# which can be obtained at http://www.apache.org/licenses/LICENSE-2.0.
# In applying this licence, ECMWF does not waive the privileges and immunities
# granted to it by virtue of its status as an intergovernmental organisation
# nor does it submit to any jurisdiction.
#

from climetlab import load_source, plot_map


def test_netcdf():
    for s in load_source("file", "docs/examples/test.nc"):
        plot_map(s)


if __name__ == "__main__":
    test_netcdf()
